'''Mail Antivir Gate Library
   version 0.3
   (c) 2003 Jan ONDREJ (SAL) <ondrejj@salstar.sk>

This program is a email antivir gate for postfix.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from default import *
from config import *
import os,sys,string,re
import pwd,grp,socket,rfc822
import time
import cStringIO

from avlib import *

TRUE=1
FALSE=0

R_OK=FALSE
R_FAIL=TRUE

S_OK=1
S_TEMPFAIL=2
S_REJECT='MAIL REJECTED'
S_DROP='MAIL DROPPED'
S_ALWAYS_SEND='SENT, virpattern matched in ALWAYS_SEND'

RECV_SMTP=0
RECV_HEADER=1
RECV_BODY=2

class mail_class:
  addr='' # Connection address
  comm='' # SMTP communication
  header='' # Email Header
  data='' # Email data
  xheader='' # Extended header
  recip=[] # recipient emails
  def __init__(self):
      self.hf=cStringIO.StringIO()
      self.df=cStringIO.StringIO()
  def close(self):
      self.hf.flush()
      self.df.flush()
      mail.header=self.hf.getvalue()
      self.hf.close()
      mail.data=mail.header+self.df.getvalue()
      self.df.close()
  def addhdr(self,desc,value=''):
      if value:
        self.xheader=self.xheader+desc+': '+value+'\r\n'
      else:
        self.xheader=self.xheader+desc
  def hdrlen(self):
      return len(self.header)
  def getnamebyaddr(self,ip):
      try:
        return socket.gethostbyaddr(ip)[0]
      except:
        return ''

mail=mail_class()
set_debug_level(DEBUG_LEVEL)

def ctime():
    return time.time()

def dochroot(user,group,newroot):
    def func_err(fn,f,arg):
        try:
          f(arg)
          return 0
        except OSError,(ec,es):
          debug(0,"WARNING! Can't do %10s: %s" % (fn,es))
        except AttributeError,es:
          debug(0,"WARNING! Can't do "+es+": Your python don't have this function!")
        return 1
    uid,gid=0,0
    if user:
      try:
        uid=pwd.getpwnam(user)[2]
      except KeyError,es:
        debug(0,"ERROR, getpwnam: ",[[user]],es)
        sys.exit(100)
    if group:
      try:
        gid=grp.getgrnam(group)[2]
      except KeyError,es:
        debug(0,"ERROR, getgrnam: ",[[group]],es)
        sys.exit(100)
    if newroot:
      if func_err('chroot',os.chroot,newroot):
        set_root_path(newroot)
    if gid:
      func_err('setgroups',os.setgroups,[gid])
      func_err('setgid',os.setgid,gid)
    if uid:
      func_err('setuid',os.setuid,uid)
    if (os.getuid()==0) or (os.getgid()==0
        or (os.geteuid()==0) or (os.getegid()==0)):
      debug(0,"ERROR! I can't setuid and i am running as root!")
      sys.exit(10)

class smtp:
  f=FALSE
  bufsize=10240
  def send(self,s):
    self.conn.send(s)

class smtps(smtp):
  def __init__(self,conn=0,addr=['stdin']):
    if conn:
      self.f=conn.makefile('rw',self.bufsize)
      self.write=conn.send
      self.readline=self.f.readline
    else:
      self.write=self.send
      self.readline=sys.stdin.readline
    self.FROM=''
    self.TO=[]
    self.RECV=RECV_SMTP
    self.temp=''
    self.t0=ctime()
    self.srv=smtpc()
    self.write(self.srv.readline())
    mail.addr='NOOP CONNECT FROM: '+addr[0]+':'+str(addr[1])\
      +', '+mail.getnamebyaddr(addr[0])+'\r\n'
  def send(self,s):
    sys.stdout.write(s)
    sys.stdout.flush()
  def send_ok(self,r="Ok"):
    self.write("250 "+r+"\r\n")
  def quit_connect(self):
    self.srv.close()
    while 1:
      line=self.readline()
      if not line: return 1
      if re.search("^QUIT",line,re.IGNORECASE):
        self.write("221 Bye\r\n")
        self.f.close()
        return 1
      else:
        self.write("502 Error: command not implemented\r\n")
        return 0
  def recv(self):
    global mail
    line=self.readline()
    if not line: return 1
    if self.RECV: # receiving data or header
      if ((self.RECV==RECV_HEADER) & ((line=='\r\n') | (line=='\n'))):
        debug(3,"SMTPS: HEADER DONE")
        mail.df.write(line)
        self.RECV=RECV_BODY
      elif (line=='.\r\n') | (line=='.\n'):
        mail.close()
        if len(mail.data)==len(mail.header):
          debug(0,"SMTPS: Warning: No header?")
          mail.header=''
        debug(3,"SMTPS: BODY DONE, size: ",len(mail.data)," B")
        self.RECV=RECV_SMTP
        v,virname=checkvir(self.FROM,self.TO)
        debug(2,"SMTPS: Total time: ",[[ctime()-self.t0]])
        if (v==S_OK) | (v==S_ALWAYS_SEND):
          debug(4,"SMTPS: Sending data")
          self.srv.conn.send(mail.header)
          self.srv.conn.send(mail.xheader)
          self.srv.conn.send(mail.data[len(mail.header):])
          debug(1,"SMTPS: OK: 250 Ok")
          self.srv.conn.send(line) # send .
          self.write(self.srv.readline())
        elif v==S_REJECT:
          debug(1,"SMTPS: REJECT: ","550 Content rejected - VIRUS ",virname)
          self.write("550 Content rejected - VIRUS "+str(virname)+"\r\n")
          return self.quit_connect()
        elif v==S_DROP:
          debug(1,"SMTPS: DROP: 250 mail dropped, VIRUS ",virname)
          self.write("250 mail dropped, VIRUS "+str(virname)+"\r\n")
          return self.quit_connect()
        else: # S_TEMPFAIL
          debug(1,"SMTPS: TEMPFAIL: 451 ",virname)
          self.write("451 "+str(virname)+"\r\n")
          return self.quit_connect()
      else:
        if self.RECV==RECV_HEADER:
          mail.hf.write(line)
        else:
          mail.df.write(line)
    else:
      debug(2,'SMTPS: '+string.rstrip(line))
      mail.comm=mail.comm+line
      mailfrom=re.search('^MAIL *FROM:(.*)$',line,re.IGNORECASE)
      rcptto=re.search('^RCPT *TO:(.*)$',line,re.IGNORECASE)
      if mailfrom:
        self.FROM=mailfrom.group(1)
      elif rcptto:
        self.TO.append(rcptto.group(1))
        try:
          mail.recip.append(rfc822.parseaddr(rcptto.group(1))[1])
        except:
          pass
      elif re.search('^DATA',line,re.IGNORECASE):
        self.RECV=RECV_HEADER
      elif re.search('^QUIT',line,re.IGNORECASE):
        self.srv.conn.send(line)
        self.write(self.srv.readline()) # 221 Bye
        return 1
      #if re.search("^[LE]HLO ",line,re.IGNORECASE): # disabling EHLO
      #  self.write("502 Error: command not implemented\r\n")
      #  return 0
      self.srv.conn.send(line)
      self.write(self.srv.conn.recv(BUFSIZE))
      return 0

class SmtpcError(Exception):
  """SMTP Client Error"""
  ConnectError='ConnectError'
  WrongReturnCode='WrongReturnCode'
  SendmailError='SendmailError'

class smtpc(smtp):
  def __init__(self):
    try:
      self.conn=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      self.conn.connect((SMTP_HOST,SMTP_PORT))
    except socket.error,(ec,es):
      debug(0,"SMTPC: ERROR: connect: ",es)
      raise SmtpcError,(SmtpcError.ConnectError,"451 SMTP connection refused\r\n")
    self.f=self.conn.makefile('rw',self.bufsize)
  def send(self,data,rc='...'):
    if data!='':
      debug(4,"SMTPC>: ",string.rstrip(data))
      self.conn.send(data+"\r\n")
    ret=''
    while 1:
      r=self.readline()
      ret=ret+r
      if re.search("^[0-9][0-9][0-9] ",r):
        break
    debug(4,"SMTPC<: ",string.rstrip(ret))
    if not re.search("^"+rc+" ",ret):
      raise SmtpcError,(SmtpcError.WrongReturnCode,string.rstrip(ret)+"!="+rc)
    return ret
  def readline(self):
    return self.f.readline()
  def close(self):
    self.f.close()
    self.conn.shutdown(2)
    self.conn.close()
  def sendmail(self,sender,recipients,ldata='',file=''):
    if file:
      ff=open(file,'r')
      firstline=ff.readline()
      ldata=ff.read()
      ff.close()
    ret="451 SMTP connection error\r\n"
    try:
      self.send('','220')
      self.send("HELO localhost",'250')
      self.send("MAIL FROM:"+sender,'250')
      for rec in recipients:
        self.send("RCPT TO:"+rec,'250')
      self.send("DATA",'354')
      if ldata:
        self.conn.send(ldata)
      else:
        self.conn.send(mail.header)
        self.conn.send(mail.xheader)
        self.conn.send(mail.data[mail.hdrlen():])
      debug(4,"SMTPC>: DATA SENT")
      ret=self.send(".","250")
    except SmtpcError.WrongReturnCode,ret:
      raise SmtpcError,(SmtpcError.SendmailError,ret)
    try:
      self.send("QUIT","221")
      self.close()
    except SmtpcError.WrongReturnCode,ret:
      debug(4,"SMTPC: quit error",ret)
    return ret

def do_scan(scnr,fn=''):
    detected,virlist='',[]
    try:
      if fn:
        scnr.filename=fn
      detected,virlist=scnr.scanstream(mail.data)
      scnr.destroy()
      return detected,virlist,''
    except socket.error,(ec,es):
      debug_traceback(4,"do_scan: ")
      if TEMPFAIL_ON_SCANNER_FAIL:
        return detected,virlist,scnr.name+": SocketError: "+es
    except (IOError,OSError):
      debug_traceback(4,"do_scan: ")
      if TEMPFAIL_ON_SCANNER_FAIL:
        return detected,virlist,scnr.name+": Error: "\
          +traceback_value_str()
    except (ScannerError,ValueError),es:
      debug_traceback(4,"do_scan: ")
      if TEMPFAIL_ON_SCANNER_FAIL:
        return detected,virlist,scnr.name+": "+str(es)

def checkvir(sender,recipients):
    try:
      detected,virlist='',[]
      t0=ctime()
      scanarr,scannames=[],'None'
      for scnr in SCANNERS:
        debug(3,"Running: ",scnr.name)
        detected,virlist,scan_reply=do_scan(scnr)
        if scan_reply:
          debug(0, "CheckVir: ERROR: ",scan_reply)
          return S_TEMPFAIL,scan_reply
        else:
          scanarr=scanarr+[scnr.name]
          scannames=string.join(scanarr,' ')
        if detected:
          debug(3,"Satus: ",[detected,virlist])
          virlist=["SCANNER: "+scnr.name+"\n"]+virlist
          break
      mail.addhdr("X-Sagator-Scanner", scannames)
      if detected=='':
        return S_OK,'250 Ok' # mail delivered OK
      else: # VIRUS FOUND
        if (QUARANTINE_DIR!='') & (DONT_QUARANTINE!='') \
            & (re.search(DONT_QUARANTINE,detected,re.IGNORECASE)!=None):
          qfname='Not quarantined, disabled in config.'
        else:
          fext=time.strftime("-%Y%m%d-%H%M%S-",time.localtime(time.time()))
          fpath=QUARANTINE_DIR+'/q'
          num=0
          while 1:
            num=num+1
            qfname=fpath+'m'+fext+str(num)
            try:
              fd=os.open(qfname,os.O_WRONLY|os.O_CREAT|os.O_EXCL, 0660)
              f=os.fdopen(fd,'w',BUFSIZE)
              f.write(mail.addr)
              f.write(mail.comm)
              f.write(mail.xheader)
              f.write(mail.data)
              f.write(".\r\nQUIT\r\n")
              f.close()
              break
            except OSError,(ec,es):
              if ec!=17: # already exists
                raise OSError,(ec,es+' '+qfname)
        mail.addhdr("X-Sagator-Status", detected)
        # SEND REPORT
        scanout=''
        for i in virlist:
          scanout=scanout+i
        rstat=S_REJECT
        if re.search(DROP,detected,re.IGNORECASE):
          rstat=S_DROP
        if (ALWAYS_SEND!='') \
          & (re.search(ALWAYS_SEND,detected,re.IGNORECASE)!=None):
            rstat=S_ALWAYS_SEND
        if ADMIN_REPORT_MSG:
          msg=ADMIN_REPORT_MSG % (
                REPORT_SENDER,string.join(REPORT_RECIPIENTS,',\r\n\t'),detected,
                mail.addr+mail.comm, rstat, qfname, scanout, mail.header)
          try:
            smtpc().sendmail(REPORT_SENDER,REPORT_RECIPIENTS,msg)
          except SmtpcError,(ses,es):
            detected=detected+'+AdminReportError/'+ses
        if USER_REPORT_MSG:
          msg=USER_REPORT_MSG % (
                REPORT_SENDER,string.join(mail.recip,',\r\n\t'),detected,
                mail.addr+mail.comm, rstat, qfname, scanout, mail.header)
          try:
            smtpc().sendmail(REPORT_SENDER,REPORT_RECIPIENTS,msg)
          except SmtpcError,(ses,es):
            detected=detected+'+UserReportError/'+ses
        return rstat,detected
    except socket.error,(ec,es):
      debug(0,"checkvir: ERROR: socket.error: ",[[ec]],es)
      debug_traceback(4,"checkvir: ")
      return S_TEMPFAIL,es
    except IOError,(ec,es):
      debug(0,"checkvir: ERROR: IO: ",es)
      debug_traceback(4,"checkvir: ")
      return S_TEMPFAIL,'IOError'
    except:
      debug_traceback(3,"checkvir: ")
      return S_TEMPFAIL,'AnyError, see sagator logs'
